import 'dart:developer';
import 'dart:io';

import 'package:dbus/dbus.dart';
import 'package:factory_reset_tools/dbus/dbus_remote_object.dart';
import 'package:path/path.dart' as p;
import 'package:retry/retry.dart';
import 'package:yaml/yaml.dart';

const defaultFilePath = '/usr/share/desktop-provision';
const defaultFileName = 'reset.yaml';

enum ResetOptionType {
  factoryReset('factory-reset'),
  fwSetup('fwsetup');

  const ResetOptionType(this.value);
  final String value;

  static ResetOptionType fromString(String value) {
    switch (value) {
      case 'factory-reset':
        return factoryReset;
      case 'fwsetup':
        return fwSetup;
      default:
        throw ArgumentError('Invalid value: $value');
    }
  }
}

sealed class BootOption {
  BootOption(this.type, this.title, this.description);
  final ResetOptionType type;
  final String title;
  final String? description;

  Future<void> run();
}

class GrubBootOption extends BootOption {
  GrubBootOption(super.key, super.title, super.description, this.optionName);
  final String optionName;

  @override
  Future<void> run() async {
    final result = await Process.run('grub-editenv', [
      '/var/lib/snapd/hostfs/boot/grub/grubenv',
      'set',
      'next_entry=$optionName',
    ]);
    stdout.write(result.stdout);
    stderr.write(result.stderr);
    if (result.exitCode != 0) {
      throw Exception(result.stderr);
    }

    final dbusClient = DBusClient.system();

    final loginObject = DBusRemoteObject(
      dbusClient,
      name: 'org.freedesktop.login1',
      path: DBusObjectPath('/org/freedesktop/login1'),
    );
    await loginObject.callMethod(
      'org.freedesktop.login1.Manager',
      'Reboot',
      [const DBusBoolean(true)],
    );
  }
}

class RunCommandBootOption extends BootOption {
  RunCommandBootOption(super.key, super.title, super.description, this.command);
  final List<String> command;

  @override
  Future<void> run() async {
    final result = await Process.run(command[0], command.sublist(1));
    stdout.write(result.stdout);
    stderr.write(result.stderr);
    if (result.exitCode == 126) {
      throw 'unauthorized';
    }
  }
}

final List<BootOption> defaultBootOption = [
  GrubBootOption(
    ResetOptionType.factoryReset,
    'Restore Ubuntu to factory state',
    'This option will restore Ubuntu to factory default, removing all files stored in this system during the process.',
    'Restore Ubuntu to factory state',
  ),
  GrubBootOption(
    ResetOptionType.fwSetup,
    'UEFI Firmware Settings',
    'Restart into UEFI Firmware (BIOS) Settings menu.',
    'UEFI Firmware Settings',
  ),
];

String _expandPathIfNull(String? path) {
  return path ??
      p.join(
        Platform.environment['DESKTOP_PROVISION_PATH'] ?? defaultFilePath,
        defaultFileName,
      );
}

List<BootOption> getResetOptions({String? path}) {
  try {
    final rawConfig = File(_expandPathIfNull(path)).readAsStringSync();
    final config =
        loadYaml(rawConfig, sourceUrl: Uri.file(_expandPathIfNull(path)));
    final options = config['reset_tool_options'] as YamlList;
    final bootOptions = <BootOption>[];
    for (final item in options) {
      if (item is! YamlMap) {
        continue;
      }
      if (item.containsKey('grub_option')) {
        bootOptions.add(
          GrubBootOption(
            ResetOptionType.fromString(item['key'] as String),
            item['title'] as String,
            item['description'] as String?,
            item['grub_option'] as String,
          ),
        );
      } else if (item.containsKey('run_command')) {
        final rawCommand = item['run_command'];
        List<String> command;
        if (rawCommand is YamlList) {
          command = rawCommand.cast<String>();
        } else {
          command = ['/usr/bin/bash', '-c', rawCommand as String];
        }

        bootOptions.add(
          RunCommandBootOption(
            ResetOptionType.fromString(item['key'] as String),
            item['title'] as String,
            item['description'] as String?,
            command,
          ),
        );
      }
    }
    assert(bootOptions.isNotEmpty);
    return bootOptions;
    // ignore: avoid_catches_without_on_clauses
  } catch (e) {
    // Generic catch-all exception
    log('reading $path for options failed', error: e);
    return defaultBootOption;
  }
}

Future<void> startCommand(
  ResetOptionType type, {
  String? path,
}) {
  BootOption option;
  // Don't have to call _expandPathIfNull() because it's
  // already called in getResetOptions
  final options = getResetOptions(path: path);

  try {
    option = options.firstWhere((option) => option.type == type);
  } on StateError {
    throw StateError('option not found');
  }

  return option.run();
}

Future<void> startCommandViaDbus(
  ResetOptionType type, {
  String? path,
}) async {
  final dbusClient = DBusClient.system();
  final object = ComCanonicalOemFactoryResetTools(
    dbusClient,
    'com.canonical.oem.FactoryResetTools',
  );

  // as service is activated by dbus call, we should try to rerun if dbus gives
  // error
  const retryRunner = RetryOptions(maxAttempts: 5);
  await retryRunner.retry(
    () => object.callReboot(type.value),
    retryIf: (e) =>
        e is DBusMethodResponseException &&
        e.errorName == 'org.freedesktop.DBus.Error.UnknownObject',
  );
}
