import 'package:flutter/material.dart';
import 'package:flutter_riverpod/flutter_riverpod.dart';
import 'package:flutter_test/flutter_test.dart';
import 'package:mockito/mockito.dart';
import 'package:ubuntu_provision/src/network/network_l10n.dart';
import 'package:ubuntu_provision/src/network/network_page.dart';
import 'package:ubuntu_provision/src/network/wifi_model.dart';
import 'package:ubuntu_provision/src/network/wifi_view.dart';
import 'package:yaru/yaru.dart';
import 'package:yaru_test/yaru_test.dart';

import 'test_network.dart';

void main() {
  testWidgets('device and access point tiles', (tester) async {
    final device1 = MockWifiDevice();
    when(device1.model).thenReturn('model1');
    when(device1.scanning).thenReturn(false);
    when(device1.isConnecting).thenReturn(false);
    when(device1.isAvailable).thenReturn(true);
    when(device1.isActive).thenReturn(false);

    final accessPoint1 = MockAccessPoint();
    when(accessPoint1.name).thenReturn('ap1');
    when(accessPoint1.strength).thenReturn(1);
    when(accessPoint1.isOpen).thenReturn(true);
    when(device1.accessPoints).thenReturn([accessPoint1]);
    when(device1.isSelectedAccessPoint(accessPoint1)).thenReturn(false);
    when(device1.isActiveAccessPoint(accessPoint1)).thenReturn(false);

    final device2 = MockWifiDevice();
    when(device2.model).thenReturn('model2');
    when(device2.scanning).thenReturn(true);
    when(device2.isConnecting).thenReturn(true);
    when(device2.isAvailable).thenReturn(true);
    when(device2.isActive).thenReturn(true);

    final accessPoint2 = MockAccessPoint();
    when(accessPoint2.name).thenReturn('ap2');
    when(accessPoint2.strength).thenReturn(2);
    when(accessPoint2.isOpen).thenReturn(false);
    when(device2.accessPoints).thenReturn([accessPoint2]);
    when(device2.isActiveAccessPoint(accessPoint2)).thenReturn(true);
    when(device2.isSelectedAccessPoint(accessPoint2)).thenReturn(true);

    final model = buildWifiModel(
      devices: [device1, device2],
      isEnabled: true,
    );

    WifiDevice? selectedDevice;
    AccessPoint? selectedAccessPoint;

    await tester.pumpApp(
      (_) => ProviderScope(
        overrides: [
          wifiModelProvider.overrideWith((_) => model),
        ],
        child: Material(
          child: Column(
            children: [
              WifiView(
                expanded: true,
                onEnabled: () {},
                onSelected: (device, accessPoint) {
                  selectedDevice = device;
                  selectedAccessPoint = accessPoint;
                },
              ),
            ],
          ),
        ),
      ),
    );

    // inactive device
    final tile1 = find.listTile('model1');
    expect(tile1, findsOneWidget);
    expect(
      find.descendant(
        of: tile1,
        matching: find.byType(YaruCircularProgressIndicator),
      ),
      findsNothing,
    );

    // connecting device
    final tile2 = find.listTile('model2');
    expect(tile2, findsOneWidget);
    expect(
      find.descendant(
        of: tile2,
        matching: find.byType(YaruCircularProgressIndicator),
      ),
      findsOneWidget,
    );

    // select ap
    final ap1 = find.listTile('ap1').first;
    expect(ap1, findsOneWidget);
    await tester.tap(ap1);
    expect(selectedDevice, equals(device1));
    expect(selectedAccessPoint, equals(accessPoint1));

    // select another ap
    final ap2 = find.listTile('ap2').first;
    expect(ap2, findsOneWidget);
    await tester.tap(ap2);
    expect(selectedDevice, equals(device2));
    expect(selectedAccessPoint, equals(accessPoint2));
  });

  testWidgets('wifi disabled', (tester) async {
    final model = buildWifiModel(isEnabled: false);

    var wasEnabled = false;

    await tester.pumpApp(
      (_) => ProviderScope(
        overrides: [
          wifiModelProvider.overrideWith((_) => model),
        ],
        child: Column(
          children: [
            WifiRadioButton(
              value: ConnectMode.wifi,
              onChanged: (v) {},
            ),
            WifiView(
              expanded: true,
              onEnabled: () => wasEnabled = true,
              onSelected: (device, accessPoint) {},
            ),
          ],
        ),
      ),
    );

    final context = tester.element(find.byType(WifiView));
    final l10n = NetworkLocalizations.of(context);

    expect(find.byType(WifiListView), findsNothing);

    final radioButtonFinder = find.byType(YaruRadioButton<ConnectMode>);
    expect(radioButtonFinder, findsOneWidget);

    final radioButton =
        tester.widget<YaruRadioButton<ConnectMode>>(radioButtonFinder);
    expect(radioButton.onChanged, isNull);

    expect(find.text(l10n.networkWifiOff), findsOneWidget);
    expect(find.text(l10n.networkWifiDisabled), findsOneWidget);

    final button = find.button(l10n.networkWifiEnable);
    expect(button, findsOneWidget);
    await tester.tap(button);
    expect(wasEnabled, isTrue);
  });

  testWidgets('no wifi devices', (tester) async {
    final model = buildWifiModel();

    await tester.pumpApp(
      (_) => ProviderScope(
        overrides: [
          wifiModelProvider.overrideWith((_) => model),
        ],
        child: Column(
          children: [
            WifiRadioButton(
              value: ConnectMode.wifi,
              onChanged: (v) {},
            ),
            WifiView(
              expanded: true,
              onEnabled: () {},
              onSelected: (device, accessPoint) {},
            ),
          ],
        ),
      ),
    );

    final context = tester.element(find.byType(WifiView));
    final l10n = NetworkLocalizations.of(context);

    final radioButtonFinder = find.byType(YaruRadioButton<ConnectMode>);
    expect(radioButtonFinder, findsOneWidget);

    final radioButton =
        tester.widget<YaruRadioButton<ConnectMode>>(radioButtonFinder);
    expect(radioButton.onChanged, isNull);

    expect(find.text(l10n.networkWifiNone), findsOneWidget);
  });

  testWidgets('starts periodic scanning', (tester) async {
    final model = buildWifiModel();

    await tester.pumpApp(
      (_) => ProviderScope(
        overrides: [
          wifiModelProvider.overrideWith((_) => model),
        ],
        child: WifiView(
          expanded: true,
          onEnabled: () {},
          onSelected: (device, accessPoint) {},
        ),
      ),
    );

    verify(model.startPeriodicScanning()).called(1);
  });
}
